"use strict";

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var _require = require("lodash");

var some = _require.some;


module.exports = function (_ref) {
  var t = _ref.types;
  var traverse = _ref.traverse;

  var removeOrVoid = require("babel-helper-remove-or-void")(t);
  var shouldRevisit = Symbol("shouldRevisit");

  var main = {
    // remove side effectless statement
    ExpressionStatement: function ExpressionStatement(path) {
      if (path.get("expression").isPure()) {
        removeOrVoid(path);
      }
    },


    Function: {

      // Let's take all the vars in a function that are not in the top level scope and hoist them
      // with the first var declaration in the top-level scope. This transform in itself may
      // not yield much returns (or even can be marginally harmful to size). However it's great
      // for taking away statements from blocks that can be only expressions which the `simplify`
      // plugin can turn into other things (e.g. if => conditional).
      exit: function exit(path) {
        // This hurts gzip size.
        if (!this.optimizeRawSize) {
          return;
        }

        var node = path.node;
        var scope = path.scope;

        var seen = new Set();
        var declars = [];
        var mutations = [];

        var _loop = function _loop(name) {
          var binding = scope.bindings[name];
          if (!binding.path.isVariableDeclarator()) {
            return "continue";
          }

          var declarPath = binding.path.parentPath;
          if (seen.has(declarPath)) {
            return "continue";
          }
          seen.add(declarPath);

          if (declarPath.parentPath.isForInStatement()) {
            return "continue";
          }

          if (declarPath.parentPath.parentPath.isFunction()) {
            return "continue";
          }

          if (!declarPath.node || !declarPath.node.declarations) {
            return "continue";
          }

          var assignmentSequence = [];
          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            var _loop2 = function _loop2() {
              var declar = _step2.value;

              declars.push(declar);
              if (declar.init) {
                assignmentSequence.push(t.assignmentExpression("=", declar.id, declar.init));
                mutations.push(function () {
                  declar.init = null;
                });
              }
            };

            for (var _iterator2 = declarPath.node.declarations[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              _loop2();
            }
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }

          if (assignmentSequence.length) {
            mutations.push(function () {
              return declarPath.replaceWith(t.sequenceExpression(assignmentSequence));
            });
          } else {
            mutations.push(function () {
              return removeOrVoid(declarPath);
            });
          }
        };

        for (var name in scope.bindings) {
          var _ret = _loop(name);

          if (_ret === "continue") continue;
        }

        if (declars.length) {
          mutations.forEach(function (f) {
            return f();
          });
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
            for (var _iterator = node.body.body[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              var statement = _step.value;

              if (t.isVariableDeclaration(statement)) {
                var _statement$declaratio;

                (_statement$declaratio = statement.declarations).push.apply(_statement$declaratio, declars);
                return;
              }
            }
          } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
              }
            } finally {
              if (_didIteratorError) {
                throw _iteratorError;
              }
            }
          }

          var varDecl = t.variableDeclaration("var", declars);
          node.body.body.unshift(varDecl);
        }
      }
    },

    // Remove bindings with no references.
    Scope: {
      exit: function exit(path) {
        if (path.node[shouldRevisit]) {
          delete path.node[shouldRevisit];
          path.visit();
        }
      },
      enter: function enter(path) {
        var _this = this;

        if (path.isProgram()) {
          return;
        }

        var scope = path.scope;

        var _loop3 = function _loop3(name) {
          var binding = scope.bindings[name];
          if (!binding.referenced && binding.kind !== "param" && binding.kind !== "module") {
            var _ret4 = function () {
              if (binding.path.isVariableDeclarator()) {
                if (binding.path.parentPath.parentPath && binding.path.parentPath.parentPath.isForInStatement()) {
                  // Can't remove if in a for in statement `for (var x in wat)`.
                  return {
                    v: "continue"
                  };
                }
              } else if (!scope.isPure(binding.path.node)) {
                return {
                  v: "continue"
                };
              } else if (binding.path.isFunctionExpression() || binding.path.isClassExpression()) {
                // `bar(function foo() {})` foo is not referenced but it's used.
                return {
                  v: "continue"
                };
              }

              var mutations = [];
              var bail = false;
              // Make sure none of the assignments value is used
              binding.constantViolations.forEach(function (p) {
                if (bail || p === binding.path) {
                  return;
                }

                if (!p.parentPath.isExpressionStatement()) {
                  bail = true;
                }

                if (p.isAssignmentExpression() && !p.get("right").isPure()) {
                  mutations.push(function () {
                    return p.replaceWith(p.get("right"));
                  });
                } else {
                  mutations.push(function () {
                    return removeOrVoid(p);
                  });
                }
              });

              if (bail) {
                return {
                  v: "continue"
                };
              }

              if (binding.path.isVariableDeclarator() && binding.path.node.init && !scope.isPure(binding.path.node.init) && binding.path.parentPath.node.declarations) {
                if (binding.path.parentPath.node.declarations.length !== 1) {
                  return {
                    v: "continue"
                  };
                }

                binding.path.parentPath.replaceWith(binding.path.node.init);
              } else {
                updateReferences(binding.path, _this);
                removeOrVoid(binding.path);
              }

              mutations.forEach(function (f) {
                return f();
              });
              scope.removeBinding(name);
            }();

            if ((typeof _ret4 === "undefined" ? "undefined" : _typeof(_ret4)) === "object") return _ret4.v;
          } else if (binding.constant) {
            if (binding.path.isFunctionDeclaration() || binding.path.isVariableDeclarator() && binding.path.get("init").isFunction()) {
              var _ret5 = function () {
                var fun = binding.path.isFunctionDeclaration() ? binding.path : binding.path.get("init");
                var allInside = true;
                var _iteratorNormalCompletion3 = true;
                var _didIteratorError3 = false;
                var _iteratorError3 = undefined;

                try {
                  for (var _iterator3 = binding.referencePaths[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                    var ref = _step3.value;

                    if (!ref.find(function (p) {
                      return p.node === fun.node;
                    })) {
                      allInside = false;
                      break;
                    }
                  }
                } catch (err) {
                  _didIteratorError3 = true;
                  _iteratorError3 = err;
                } finally {
                  try {
                    if (!_iteratorNormalCompletion3 && _iterator3.return) {
                      _iterator3.return();
                    }
                  } finally {
                    if (_didIteratorError3) {
                      throw _iteratorError3;
                    }
                  }
                }

                if (allInside) {
                  scope.removeBinding(name);
                  updateReferences(binding.path, _this);
                  removeOrVoid(binding.path);
                  return {
                    v: "continue"
                  };
                }
              }();

              if ((typeof _ret5 === "undefined" ? "undefined" : _typeof(_ret5)) === "object") return _ret5.v;
            }

            if (binding.references === 1 && binding.kind !== "param" && binding.kind !== "module" && binding.constant) {
              var _ret6 = function () {
                var replacement = binding.path.node;
                var replacementPath = binding.path;
                if (t.isVariableDeclarator(replacement)) {
                  replacement = replacement.init;
                  replacementPath = replacementPath.get("init");
                }
                if (!replacement) {
                  return {
                    v: "continue"
                  };
                }

                if (!scope.isPure(replacement, true)) {
                  return {
                    v: "continue"
                  };
                }

                if (binding.referencePaths.length > 1) {
                  throw new Error("Expected only one reference");
                }

                var bail = false;
                var refPath = binding.referencePaths[0];

                if (replacementPath.isIdentifier()) {
                  bail = refPath.scope.getBinding(replacement.name) !== scope.getBinding(replacement.name);
                } else {
                  replacementPath.traverse({
                    Function: function Function(path) {
                      path.skip();
                    },
                    ReferencedIdentifier: function ReferencedIdentifier(_ref2) {
                      var node = _ref2.node;

                      if (bail) {
                        return;
                      }
                      bail = refPath.scope.getBinding(node.name) !== scope.getBinding(node.name);
                    }
                  });
                }

                if (bail) {
                  return {
                    v: "continue"
                  };
                }

                var parent = binding.path.parent;
                if (t.isVariableDeclaration(parent)) {
                  parent = binding.path.parentPath.parent;
                }

                // 1. Make sure we share the parent with the node. In other words it's lexically defined
                // and not in an if statement or otherwise.
                // 2. If the replacement is an object then we have to make sure we are not in a loop or a function
                // because otherwise we'll be inlining and doing a lot more allocation than we have to
                // which would also could affect correctness in that they are not the same reference.
                var mayLoop = false;
                var sharesRoot = refPath.find(function (_ref3) {
                  var node = _ref3.node;

                  if (!mayLoop) {
                    mayLoop = t.isWhileStatement(node) || t.isFor(node) || t.isFunction(node);
                  }
                  return node === parent;
                });

                // Anything that inherits from Object.
                var isObj = function isObj(n) {
                  return t.isFunction(n) || t.isObjectExpression(n) || t.isArrayExpression(n);
                };
                var isReplacementObj = isObj(replacement) || some(replacement, isObj);

                if (!sharesRoot || isReplacementObj && mayLoop) {
                  return {
                    v: "continue"
                  };
                }

                var replaced = replace(binding.referencePaths[0], {
                  binding: binding,
                  scope: scope,
                  replacement: replacement
                });

                if (replaced) {
                  scope.removeBinding(name);
                  if (binding.path.node) {
                    removeOrVoid(binding.path);
                  }
                }
              }();

              if ((typeof _ret6 === "undefined" ? "undefined" : _typeof(_ret6)) === "object") return _ret6.v;
            }
          }
        };

        for (var name in scope.bindings) {
          var _ret3 = _loop3(name);

          if (_ret3 === "continue") continue;
        }
      }
    },

    // Remove unreachable code.
    BlockStatement: function BlockStatement(path) {
      var paths = path.get("body");

      var purge = false;

      for (var i = 0; i < paths.length; i++) {
        var p = paths[i];

        if (!purge && p.isCompletionStatement()) {
          purge = true;
          continue;
        }

        if (purge && !p.isFunctionDeclaration()) {
          removeOrVoid(p);
        }
      }
    },


    // Double check unreachable code and remove return statements that
    // have no semantic meaning
    ReturnStatement: function ReturnStatement(path) {
      var node = path.node;

      if (!path.inList) {
        return;
      }

      // Not last in it's block? (See BlockStatement visitor)
      if (path.container.length - 1 !== path.key && !path.getSibling(path.key + 1).isFunctionDeclaration() && path.parentPath.isBlockStatement()) {
        // This is probably a new oppurtinity by some other transform
        // let's call the block visitor on this again before proceeding.
        path.parentPath.pushContext(path.context);
        path.parentPath.visit();
        path.parentPath.popContext();

        return;
      }

      if (node.argument) {
        return;
      }

      var noNext = true;
      var parentPath = path.parentPath;
      while (parentPath && !parentPath.isFunction() && noNext) {
        var nextPath = parentPath.getSibling(parentPath.key + 1);
        if (nextPath.node) {
          if (nextPath.isReturnStatement()) {
            nextPath.pushContext(path.context);
            nextPath.visit();
            nextPath.popContext();
            if (parentPath.getSibling(parentPath.key + 1).node) {
              noNext = false;
              break;
            }
          } else {
            noNext = false;
            break;
          }
        }

        parentPath = parentPath.parentPath;
      }

      if (noNext) {
        removeOrVoid(path);
      }
    },
    ConditionalExpression: function ConditionalExpression(path) {
      var node = path.node;

      var evaluateTest = path.get("test").evaluateTruthy();
      if (evaluateTest === true) {
        path.replaceWith(node.consequent);
      } else if (evaluateTest === false) {
        path.replaceWith(node.alternate);
      }
    },


    IfStatement: {
      exit: function exit(path) {
        var consequent = path.get("consequent");
        var alternate = path.get("alternate");
        var test = path.get("test");

        var evaluateTest = test.evaluateTruthy();

        // we can check if a test will be truthy 100% and if so then we can inline
        // the consequent and completely ignore the alternate
        //
        //   if (true) { foo; } -> { foo; }
        //   if ("foo") { foo; } -> { foo; }
        //
        if (evaluateTest === true) {
          path.replaceWithMultiple([].concat(_toConsumableArray(toStatements(consequent)), _toConsumableArray(extractVars(alternate))));
          return;
        }

        // we can check if a test will be falsy 100% and if so we can inline the
        // alternate if there is one and completely remove the consequent
        //
        //   if ("") { bar; } else { foo; } -> { foo; }
        //   if ("") { bar; } ->
        //
        if (evaluateTest === false) {
          if (alternate.node) {
            path.replaceWithMultiple([].concat(_toConsumableArray(toStatements(alternate)), _toConsumableArray(extractVars(consequent))));
            return;
          } else {
            path.replaceWithMultiple(extractVars(consequent));
          }
        }

        // remove alternate blocks that are empty
        //
        //   if (foo) { foo; } else {} -> if (foo) { foo; }
        //
        if (alternate.isBlockStatement() && !alternate.node.body.length) {
          alternate.remove();
          // For if-statements babel-traverse replaces with an empty block
          path.node.alternate = null;
        }

        // if the consequent block is empty turn alternate blocks into a consequent
        // and flip the test
        //
        //   if (foo) {} else { bar; } -> if (!foo) { bar; }
        //
        if (consequent.isBlockStatement() && !consequent.node.body.length && alternate.isBlockStatement() && alternate.node.body.length) {
          consequent.replaceWith(alternate.node);
          alternate.remove();
          // For if-statements babel-traverse replaces with an empty block
          path.node.alternate = null;
          test.replaceWith(t.unaryExpression("!", test.node, true));
        }
      }
    },

    SwitchStatement: {
      exit: function exit(path) {
        var evaluated = path.get("discriminant").evaluate();

        if (!evaluated.confident) return;

        var discriminant = evaluated.value;
        var cases = path.get("cases");

        var matchingCaseIndex = -1;
        var defaultCaseIndex = -1;

        for (var i = 0; i < cases.length; i++) {
          var test = cases[i].get("test");

          // handle default case
          if (test.node === null) {
            defaultCaseIndex = i;
            continue;
          }

          var testResult = test.evaluate();

          // if we are not able to deternine a test during
          // compile time, we terminate immediately
          if (!testResult.confident) return;

          if (testResult.value === discriminant) {
            matchingCaseIndex = i;
            break;
          }
        }

        var result = void 0;

        if (matchingCaseIndex === -1) {
          if (defaultCaseIndex === -1) {
            path.skip();
            path.replaceWithMultiple(extractVars(path));
            return;
          } else {
            result = getStatementsUntilBreak(defaultCaseIndex);
          }
        } else {
          result = getStatementsUntilBreak(matchingCaseIndex);
        }

        if (result.bail) return;

        // we extract vars from the entire switch statement
        // and there will be duplicates which
        // will be again removed by DCE
        replaceSwitch([].concat(_toConsumableArray(extractVars(path)), _toConsumableArray(result.statements)));

        function getStatementsUntilBreak(start) {
          var result = {
            bail: false,
            statements: []
          };

          for (var _i = start; _i < cases.length; _i++) {
            var consequent = cases[_i].get("consequent");

            for (var j = 0; j < consequent.length; j++) {
              var _isBreaking = isBreaking(consequent[j], path);
              if (_isBreaking.bail) {
                result.bail = true;
                return result;
              }
              if (_isBreaking.break) {
                // compute no more
                // exit out of the loop
                return result;
              } else {
                result.statements.push(consequent[j].node);
              }
            }
          }

          return result;
        }

        function replaceSwitch(statements) {
          var isBlockRequired = false;

          for (var _i2 = 0; _i2 < statements.length; _i2++) {
            if (t.isVariableDeclaration(statements[_i2], { kind: "let" })) {
              isBlockRequired = true;
              break;
            }
            if (t.isVariableDeclaration(statements[_i2], { kind: "const" })) {
              isBlockRequired = true;
              break;
            }
          }

          if (isBlockRequired) {
            path.replaceWith(t.BlockStatement(statements));
          } else {
            path.replaceWithMultiple(statements);
          }
        }
      }
    },

    WhileStatement: function WhileStatement(path) {
      var test = path.get("test");
      var result = test.evaluate();
      if (result.confident && !result.value) {
        path.remove();
      }
    },
    ForStatement: function ForStatement(path) {
      var test = path.get("test");
      var result = test.evaluate();
      if (result.confident) {
        if (result.value) {
          test.remove();
        } else {
          path.remove();
        }
      }
    },
    DoWhileStatement: function DoWhileStatement(path) {
      var test = path.get("test");
      var result = test.evaluate();
      if (result.confident && !result.value) {
        path.replaceWith(path.get("body").node);
      }
    },


    // Join assignment and definition when in sequence.
    // var x; x = 1; -> var x = 1;
    AssignmentExpression: function AssignmentExpression(path) {
      if (!path.get("left").isIdentifier() || !path.parentPath.isExpressionStatement()) {
        return;
      }

      var prev = path.parentPath.getSibling(path.parentPath.key - 1);
      if (!(prev && prev.isVariableDeclaration())) {
        return;
      }

      var declars = prev.node.declarations;
      if (declars.length !== 1 || declars[0].init || declars[0].id.name !== path.get("left").node.name) {
        return;
      }
      declars[0].init = path.node.right;
      removeOrVoid(path);
    },


    // Remove named function expression name. While this is dangerous as it changes
    // `function.name` all minifiers do it and hence became a standard.
    "FunctionExpression|ClassExpression": function FunctionExpressionClassExpression(path) {
      if (!this.keepFnames) {
        removeUnreferencedId(path);
      }
    },


    // Put the `var` in the left if feasible.
    ForInStatement: function ForInStatement(path) {
      var left = path.get("left");
      if (!left.isIdentifier()) {
        return;
      }

      var binding = path.scope.getBinding(left.node.name);
      if (!binding) {
        return;
      }

      if (binding.scope.getFunctionParent() !== path.scope.getFunctionParent()) {
        return;
      }

      if (!binding.path.isVariableDeclarator()) {
        return;
      }

      if (binding.path.parentPath.parentPath.isForInStatement({ left: binding.path.parent })) {
        return;
      }

      // If it has company then it's probably more efficient to keep.
      if (binding.path.parent.declarations.length > 1) {
        return;
      }

      // meh
      if (binding.path.node.init) {
        return;
      }

      removeOrVoid(binding.path);
      path.node.left = t.variableDeclaration("var", [t.variableDeclarator(left.node)]);
      binding.path = path.get("left").get("declarations")[0];
    }
  };

  return {
    name: "minify-dead-code-elimination",
    visitor: {
      Program: function Program(path) {
        var _ref4 = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

        var _ref4$opts = _ref4.opts;
        _ref4$opts = _ref4$opts === undefined ? {} : _ref4$opts;
        var _ref4$opts$optimizeRa = _ref4$opts.optimizeRawSize;
        var optimizeRawSize = _ref4$opts$optimizeRa === undefined ? false : _ref4$opts$optimizeRa;
        var _ref4$opts$keepFnames = _ref4$opts.keepFnames;
        var keepFnames = _ref4$opts$keepFnames === undefined ? false : _ref4$opts$keepFnames;

        // We need to run this plugin in isolation.
        path.traverse(main, {
          functionToBindings: new Map(),
          optimizeRawSize: optimizeRawSize,
          keepFnames: keepFnames
        });
      }
    }
  };

  function toStatements(path) {
    var node = path.node;

    if (path.isBlockStatement()) {
      var hasBlockScoped = false;

      for (var i = 0; i < node.body.length; i++) {
        var bodyNode = node.body[i];
        if (t.isBlockScoped(bodyNode)) {
          hasBlockScoped = true;
        }
      }

      if (!hasBlockScoped) {
        return node.body;
      }
    }
    return [node];
  }

  // Extracts vars from a path
  // Useful for removing blocks or paths that can contain
  // variable declarations inside them
  // Note:
  // drops are inits
  // extractVars({ var x = 5, y = x }) => var x, y;
  function extractVars(path) {
    var declarators = [];

    if (path.isVariableDeclaration({ kind: "var" })) {
      var _iteratorNormalCompletion4 = true;
      var _didIteratorError4 = false;
      var _iteratorError4 = undefined;

      try {
        for (var _iterator4 = path.node.declarations[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
          var decl = _step4.value;

          declarators.push(t.variableDeclarator(decl.id));
        }
      } catch (err) {
        _didIteratorError4 = true;
        _iteratorError4 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion4 && _iterator4.return) {
            _iterator4.return();
          }
        } finally {
          if (_didIteratorError4) {
            throw _iteratorError4;
          }
        }
      }
    } else {
      path.traverse({
        VariableDeclaration: function VariableDeclaration(varPath) {
          if (!varPath.isVariableDeclaration({ kind: "var" })) return;
          if (!isSameFunctionScope(varPath, path)) return;

          var _iteratorNormalCompletion5 = true;
          var _didIteratorError5 = false;
          var _iteratorError5 = undefined;

          try {
            for (var _iterator5 = varPath.node.declarations[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
              var _decl = _step5.value;

              declarators.push(t.variableDeclarator(_decl.id));
            }
          } catch (err) {
            _didIteratorError5 = true;
            _iteratorError5 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion5 && _iterator5.return) {
                _iterator5.return();
              }
            } finally {
              if (_didIteratorError5) {
                throw _iteratorError5;
              }
            }
          }
        }
      });
    }

    if (declarators.length <= 0) return [];

    return [t.variableDeclaration("var", declarators)];
  }

  function replace(path, options) {
    var replacement = options.replacement;
    var scope = options.scope;
    var binding = options.binding;

    // Same name, different binding.

    if (scope.getBinding(path.node.name) !== binding) {
      return;
    }

    // We don't want to move code around to different scopes because:
    // 1. Original bindings that is referenced could be shadowed
    // 2. Moving defintions to potentially hot code is bad
    if (scope !== path.scope) {
      var _ret7 = function () {
        if (t.isClass(replacement) || t.isFunction(replacement)) {
          return {
            v: void 0
          };
        }

        var bail = false;
        traverse(replacement, {
          Function: function Function(path) {
            if (bail) {
              return;
            }
            bail = true;
            path.stop();
          }
        }, scope);

        if (bail) {
          return {
            v: void 0
          };
        }
      }();

      if ((typeof _ret7 === "undefined" ? "undefined" : _typeof(_ret7)) === "object") return _ret7.v;
    }

    // Avoid recursion.
    if (path.find(function (_ref5) {
      var node = _ref5.node;
      return node === replacement;
    })) {
      return;
    }

    // https://github.com/babel/babili/issues/130
    if (!t.isExpression(replacement)) {
      t.toExpression(replacement);
    }

    // We don't remove fn name here, we let the FnExpr & ClassExpr visitors
    // check its references and remove unreferenced ones
    // if (t.isFunction(replacement)) {
    //   replacement.id = null;
    // }

    path.replaceWith(replacement);
    return true;
  }

  function updateReferences(fnToDeletePath) {
    if (!fnToDeletePath.isFunction()) {
      return;
    }

    fnToDeletePath.traverse({
      ReferencedIdentifier: function ReferencedIdentifier(path) {
        var node = path.node;
        var scope = path.scope;

        var binding = scope.getBinding(node.name);

        if (!binding || !binding.path.isFunction() || binding.scope === scope || !binding.constant) {
          return;
        }

        var index = binding.referencePaths.indexOf(path);
        if (index === -1) {
          return;
        }
        binding.references--;
        binding.referencePaths.splice(index, 1);
        if (binding.references === 0) {
          binding.referenced = false;
        }

        if (binding.references <= 1 && binding.scope.path.node) {
          binding.scope.path.node[shouldRevisit] = true;
        }
      }
    });
  }

  function removeUnreferencedId(path) {
    var id = path.get("id").node;
    if (!id) {
      return;
    }

    var node = path.node;
    var scope = path.scope;

    var binding = scope.getBinding(id.name);

    // Check if shadowed or is not referenced.
    if (binding.path.node !== node || !binding.referenced) {
      node.id = null;
    }
  }

  // path1 -> path2
  // is path1 an ancestor of path2
  function isAncestor(path1, path2) {
    return !!path2.findParent(function (parent) {
      return parent === path1;
    });
  }

  function isSameFunctionScope(path1, path2) {
    return path1.scope.getFunctionParent() === path2.scope.getFunctionParent();
  }

  // tells if a "stmt" is a break statement that would break the "path"
  function isBreaking(stmt, path) {
    if (stmt.isBreakStatement()) {
      return _isBreaking(stmt, path);
    }

    var isBroken = false;
    var result = {
      break: false,
      bail: false
    };

    stmt.traverse({
      BreakStatement: function BreakStatement(breakPath) {
        // if we already detected a break statement,
        if (isBroken) return;

        result = _isBreaking(breakPath, path);

        if (result.bail || result.break) {
          isBroken = true;
        }
      }
    });

    return result;

    function _isBreaking(breakPath, path) {
      var label = breakPath.get("label");

      if (label.node !== null) {
        // labels are fn scoped and not accessible by inner functions
        // path is the switch statement
        if (!isSameFunctionScope(path, breakPath)) {
          // we don't have to worry about this break statement
          return {
            break: false,
            bail: false
          };
        }

        // here we handle the break labels
        // if they are outside switch, we bail out
        // if they are within the case, we keep them
        var _isAncestor = isAncestor(path.scope.getBinding(label.node.name).path, path);

        return {
          bail: _isAncestor,
          break: _isAncestor
        };
      }

      // set the flag that it is indeed breaking
      var isBreak = true;

      // this flag is to capture
      // switch(0) { case 0: while(1) if (x) break; }
      var possibleRunTimeBreak = false;

      // and compute if it's breaking the correct thing
      var parent = breakPath.parentPath;

      while (parent !== stmt.parentPath) {
        // loops and nested switch cases
        if (parent.isLoop() || parent.isSwitchCase()) {
          // invalidate all the possible runtime breaks captured
          // while (1) { if (x) break; }
          possibleRunTimeBreak = false;

          // and set that it's not breaking our switch statement
          isBreak = false;
          break;
        }
        //
        // this is a special case and depends on
        // the fact that SwitchStatement is handled in the
        // exit hook of the traverse
        //
        // switch (0) {
        //   case 0: if (x) break;
        // }
        //
        // here `x` is runtime only.
        // in this case, we need to bail out. So we depend on exit hook
        // of switch so that, it would have visited the IfStatement first
        // before the SwitchStatement and would have removed the
        // IfStatement if it was a compile time determined
        //
        if (parent.isIfStatement()) {
          possibleRunTimeBreak = true;
        }
        parent = parent.parentPath;
      }

      return {
        break: possibleRunTimeBreak || isBreak,
        bail: possibleRunTimeBreak
      };
    }
  }
};